/*------------------------------------------------
* Author : SARR Serigne Khadim
* Class : MP2I
* Purpose : Boyer-Moore pattern fletching program
* Created on : 10/06/2022
------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*--- STRUCTURES ---*/

typedef struct
{
  char letter; // Only useful for display
  int jump;
} dataP; // Structure for saving jump table's elements


/*--- PROTOTYPES --- */

dataP* initJumpTable(const char pattern[]);
void forward(int* ip, int* is, int* n, int* m);
int inPattern(char letter, const char pattern[], int init);
void displayJumpTable(dataP* jumpTable, int patternLen);
void displayComparison(const char pattern[], const char sequence[], int ip, int is);
int* boyer_moore(const char pattern[], const char sequence[], int display);

/*--- MAIN PROGRAM ---*/

/*
Tests :
- BonjourMyFriendHELLOOyBOIheheHELLO & HELLO
- BonjourMyFriendHELLOOyBOIhehe & AELLO
- MiaoumeowMIoaumiaoumiaoumiaouw & miaou
*/

int main()
{
  char pattern[] = "ours";
  char sequence[] = "J'aime les ours polaires";

  int* indices = boyer_moore(pattern, sequence, 1); // Last int for display (1:y|0:n)

  return 0;
}

/*--- FUNCTIONS ---*/

// Increase is and reset ip the right way
void forward(int* ip, int* is, int* n, int* m)
{
  *is += *n; // is forward
  *ip = *n-1; // ip init

  if(*is > *m-1) // After increasing indiceS we need to check if we are still in the sequence
    *is = *m-1;
}

// Setup the jumpTable with BOYER-MOORE specs.
dataP* initJumpTable(const char pattern[]) // Working :)
{
  int n = strlen(pattern);
  dataP* jumpTable = malloc(sizeof(dataP)*n);
  
  for(int i = 0; i < n; i++)
  {
    jumpTable[i].letter = pattern[i];
    jumpTable[i].jump = n-i-1;
  }
  
  return jumpTable;
}

// WORKING !!
// Returns the first occurence's indice of a letter in the pattern (from the right, from a defined indice called init) or -1 if there isn't occurences.
int inPattern(char letter, const char pattern[], int init)
{
  for(int i = init; i >= 0; i--)
  {
    if(letter == pattern[i])
    {
      return i;
    }
  }
  return -1;
}

// Working !!
// Displays the jump table in a fancy way
void displayJumpTable(dataP* jumpTable, int patternLen)
{
  printf("\n--- Displaying jump table ---\n");

  for (int i = 0; i < patternLen; i++)
  {
    printf("* Case %d : [ %c | %d ]\n", i, jumpTable[i].letter, jumpTable[i].jump);
  }

  printf("--- Done ---\n");
}

// Working !!
// Displays the pattern seeked and the sequence with two cursors (ip is the pattern cursor (indice), is the sequence cursor)
void displayComparison(const char pattern[], const char sequence[], int ip, int is)
{
  int n = strlen(pattern);
  int m = strlen(sequence);

  printf("\n");
  // Pattern cursor
  for(int i = 0; i < is+ip-(n-1); i++)
  {
    printf(" ");
  }
  printf("! (%d)\n", ip);

  // Pattern & sequence
  for(int i = 0; i < is-(n-1); i++)
  {
    printf(" ");
  }
  printf("%s\n", pattern);
  printf("%s\n", sequence);

  // Sequence cursors
  for(int i = 0; i < is-1; i++)
  {
    if(i == is-(n-1))
      {printf("^");}

    printf(" ");
  }
  printf("^ (%d | %d)\n", is-(n-1), is);

  // Esthetic line
  for(int i = 0; i < m; i++)
  {
    printf("-");
  }
}

int* boyer_moore(const char pattern[], const char sequence[], int display)
{
  int* iFound = (int*)calloc(0, sizeof(int)); // Dynamic table to store the indices found

  if(iFound == NULL)
  {
    printf("Calloc failed with indice table...");
    return NULL;
  }

  int n = strlen(pattern);
  int m = strlen(sequence);

  dataP* jumpTable = initJumpTable(pattern);
  if(display)
    {displayJumpTable(jumpTable, n);}
  
  int indiceP = n-1; // indice for pattern
  int indiceS = n-1; // indice for sequence
  int letterIn, tempTest;

  int patternsFound = 0;
  int end = 0;

  while(!end)
  {
    if(display)
    {
      displayComparison(pattern, sequence, indiceP, indiceS);
    }

    letterIn = inPattern(sequence[indiceS], pattern, indiceP);
    // Read description, this function actually does two (three !) things in one
    // ...Checking if the pattern contains a specific letter, then returns the indice of that letter in the pattern or -1
    // ...And finally, doing that from a "init" indice (in order to restrain the pattern for future comparisons #step5)

    tempTest = (n-1)-indiceP; // Used to compare letters of a sequence's window to a pattern using indiceP
    if(sequence[indiceS-tempTest] != pattern[indiceP]) // If the last letter of the pattern doesn't match with a specific letter of the sequence...
    {
      if(letterIn < 0) // Then if that letter isn't in the pattern...
      {
        if(indiceS == m-1) // It's useless to go further if we already reached the sequence's end
        {
          end = 1;
        }
        else
        {
          forward(&indiceP, &indiceS, &n, &m);
        }
      }
      else // Otherwise the letter have a chance to match the pattern
      {
        indiceS += jumpTable[letterIn].jump; // We increase the sequence indice using the jumpTable and inPattern result (letterIn) !

        if(indiceS > m-1) // If you jumped off a cliff... the end.
          {end = 1;}
        else if(display)
          {printf(" jumped ! >>");;}
      }
    }
    else if(indiceP == 0) // If that's the case, it means that we fully compared the motif and it totally matches 
    {
      patternsFound += 1;
      iFound = (int*)realloc(iFound, patternsFound*sizeof(int)); //Reallocating iFound to add the found indice

      if(iFound == NULL)
      {
        printf("\n\n \"iFound\" realloc failed.");
        return NULL;
      }

      iFound[patternsFound-1] = indiceS-(n-1);

      if (indiceS == m-1) // If we reached the end
      {
        end = 1;
      }
      else
      {
        forward(&indiceP, &indiceS, &n, &m);
      }
    }
    else // ...Otherwise we want to decrease indiceP to continue the comparison (because we already found some matches)
    {
      indiceP--;
    }
  }

  if(patternsFound > 0)
  {
    printf("\n\nPattern found at indice(s) : ");
    for (int i = 0; i < patternsFound; i++)
    {
      printf("{%d} ", iFound[i]);
    }
  }
  else
  {
    printf("\n\nPattern not found.");
  }

  free(jumpTable); // Hehe.
  return iFound;
}